"""
Hatch build hook for compiling CFFI extensions.

This hook runs during the build process to compile the CFFI C extensions
before the wheel is created.
"""

import os
import sys
from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    """Build hook that compiles CFFI extensions."""

    def initialize(self, version, build_data):
        """
        Initialize the build hook and compile CFFI extensions.

        This method is called by hatchling during the build process.
        """
        print("=" * 60)
        print("Building CFFI extensions for PyReduce")
        print("=" * 60)

        # Get paths
        project_root = Path(self.root)
        clib_dir = project_root / "pyreduce" / "clib"

        # Add project root to sys.path so we can import
        if str(project_root) not in sys.path:
            sys.path.insert(0, str(project_root))

        try:
            # Import the CFFI FFI builder
            from cffi import FFI

            # Build vertical extraction
            print("\n[1/2] Building vertical extraction extension...")
            ffibuilder_vertical = FFI()

            with open(clib_dir / "slit_func_bd.h") as f:
                ffibuilder_vertical.cdef(f.read())

            with open(clib_dir / "slit_func_bd.c") as f:
                ffibuilder_vertical.set_source(
                    "_slitfunc_bd",
                    f.read(),
                    include_dirs=[str(clib_dir), str(clib_dir / "Release")],
                    depends=["slit_func_bd.h"],
                )

            old_cwd = os.getcwd()
            try:
                os.chdir(clib_dir)
                ffibuilder_vertical.compile(verbose=True)
                print("✓ Vertical extraction extension built successfully\n")
            finally:
                os.chdir(old_cwd)

            # Build curved extraction
            print("[2/2] Building curved extraction extension...")
            ffibuilder_curved = FFI()

            with open(clib_dir / "slit_func_2d_xi_zeta_bd.h") as f:
                ffibuilder_curved.cdef(f.read())

            with open(clib_dir / "slit_func_2d_xi_zeta_bd.c") as f:
                ffibuilder_curved.set_source(
                    "_slitfunc_2d",
                    f.read(),
                    include_dirs=[str(clib_dir), str(clib_dir / "Release")],
                    depends=["slit_func_2d_xi_zeta_bd.h"],
                )

            old_cwd = os.getcwd()
            try:
                os.chdir(clib_dir)
                ffibuilder_curved.compile(verbose=True)
                print("✓ Curved extraction extension built successfully\n")
            finally:
                os.chdir(old_cwd)

            print("=" * 60)
            print("All CFFI extensions built successfully!")
            print("=" * 60)

        except Exception as e:
            print(f"ERROR: Failed to build CFFI extensions: {e}", file=sys.stderr)
            import traceback

            traceback.print_exc()
            raise



"""
Handles instrument specific info for the METIS_LSS LSS spectrograph

Mostly reading data from the header
"""

import logging
import os.path

from .common import InstrumentWithModes

logger = logging.getLogger(__name__)


class METIS_LSS(InstrumentWithModes):
    def add_header_info(self, header, mode, **kwargs):
        """read data from header and add it as REDUCE keyword back to the header"""
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)

        # header["e_backg"] = (
        #     header["e_readn"] + header["e_exptime"] * header["e_drk"] / 3600
        # )
        #
        # header["e_ra"] /= 15
        # if header["e_jd"] is not None:
        #     header["e_jd"] += header["e_exptime"] / 2 / 3600 / 24 + 0.5

        return header

    def get_extension(self, header, mode):
        extension = 1

        return extension

    def get_wavecal_filename(self, header, mode, **kwargs):
        """Get the filename of the wavelength calibration config file"""
        # info = self.load_info()
        cwd = os.path.dirname(__file__)
        fname = f"metis_lss_{mode.lower()}_2D.npz"
        # fname = f"metis_lss_LSS_L_2D.npz" ## f"micado_IJ_2D_det1.npz"
        fname = os.path.join(cwd, "..", "wavecal", fname)

        return fname



"""
Handles instrument specific info for the HARPN spectrograph

Mostly reading data from the header
"""

import logging
import re
from os.path import dirname, join

import numpy as np

from .common import Instrument
from .filters import Filter, InstrumentFilter, NightFilter, ObjectFilter

logger = logging.getLogger(__name__)


class TypeFilter(Filter):
    def __init__(self, keyword="TNG DPR TYPE"):
        super().__init__(keyword, regex=True)

    def classify(self, value):
        if value is not None:
            match = self.match(value)
            data = np.asarray(self.data)
            data = np.unique(data[match])
            try:
                regex = re.compile(value)
                keys = [regex.match(f) for f in data]
                keys = [[g for g in d.groups() if g is not None][0] for d in keys]
                unique = np.unique(keys)
                assign = {
                    u: [d for k, d in zip(keys, data, strict=False) if k == u]
                    for u in unique
                }
                data = [(u, self.match("|".join(a))) for u, a in assign.items()]
            except IndexError:
                data = np.asarray(self.data)
                data = np.unique(data[match])
                data = [(d, self.match(d)) for d in data]
        else:
            data = np.unique(self.data)
            data = [(d, self.match(d)) for d in data]
        return data


class HARPN(Instrument):
    def __init__(self):
        super().__init__()
        self.filters = {
            "instrument": InstrumentFilter(self.info["instrument"]),
            "night": NightFilter(self.info["date"]),
            # "branch": Filter(, regex=True),
            "mode": Filter(
                self.info["instrument_mode"], regex=True, flags=re.IGNORECASE
            ),
            "type": TypeFilter(self.info["observation_type"]),
            "target": ObjectFilter(self.info["target"], regex=True),
        }
        self.night = "night"
        self.science = "science"
        self.shared = [
            "instrument",
            "night",
            "mode",
        ]
        self.find_closest = [
            "bias",
            "flat",
            "wavecal_master",
            "freq_comb_master",
            "orders",
            "scatter",
        ]

    def get_expected_values(self, target, night, mode, fiber):
        """Determine the default expected values in the headers for a given observation configuration

        Any parameter may be None, to indicate that all values are allowed

        Parameters
        ----------
        target : str
            Name of the star / observation target
        night : str
            Observation night/nights
        Returns
        -------
        expectations: dict
            Dictionary of expected header values, with one entry per step.
            The entries for each step refer to the filters defined in self.filters

        Raises
        ------
        ValueError
            Invalid combination of parameters
        """
        if target is not None:
            target = target.replace(" ", r"(?:\s*|-)")
        else:
            target = ".*"

        id_orddef = "LAMP,DARK,TUN"
        id_spec = "STAR,WAVE"

        expectations = {
            "bias": {"instrument": "HARPN", "night": night, "type": r"BIAS,BIAS"},
            "flat": {"instrument": "HARPN", "night": night, "type": r"LAMP,LAMP,TUN"},
            "orders": {
                "instrument": "HARPN",
                "night": night,
                "type": id_orddef,
            },
            "scatter": {
                "instrument": "HARPN",
                "night": night,
                "type": id_orddef,  # Same as orders or same as flat?
            },
            "wavecal_master": {
                "instrument": "HARPN",
                "night": night,
                "type": r"WAVE,WAVE,THAR2",
            },
            "freq_comb_master": {
                "instrument": "HARPN",
                "night": night,
                "type": r"WAVE,WAVE,COMB",
            },
            "science": {
                "instrument": "HARPN",
                "night": night,
                "mode": mode,
                "type": id_spec,
                "target": target,
            },
        }
        return expectations

    def get_extension(self, header, mode):
        extension = super().get_extension(header, mode)

        try:
            if (
                header["NAXIS"] == 2
                and header["NAXIS1"] == 4296
                and header["NAXIS2"] == 4096
            ):
                extension = 0
        except KeyError:
            pass

        return extension

    def add_header_info(self, header, mode, **kwargs):
        """read data from header and add it as REDUCE keyword back to the header"""
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)

        try:
            header["e_ra"] /= 15
            header["e_jd"] += header["e_exptim"] / (7200 * 24) + 0.5

        except:
            pass

        try:
            if (
                header["NAXIS"] == 2
                and header["NAXIS1"] == 4296
                and header["NAXIS2"] == 4096
            ):
                # both modes are in the same image
                prescan_x = 50
                overscan_x = 50
                naxis_x = 2148
                if mode == "BLUE":
                    header["e_xlo"] = prescan_x
                    header["e_xhi"] = naxis_x - overscan_x
                elif mode == "RED":
                    header["e_xlo"] = naxis_x + prescan_x
                    header["e_xhi"] = 2 * naxis_x - overscan_x
        except KeyError:
            pass

        return header

    def get_wavecal_filename(self, header, mode, **kwargs):
        """Get the filename of the wavelength calibration config file"""
        cwd = dirname(__file__)
        fname = f"harpn_{mode.lower()}_2D.npz"
        fname = join(cwd, "..", "wavecal", fname)
        return fname

    def get_wavelength_range(self, header, mode, **kwargs):
        wave_range = super().get_wavelength_range(header, mode, **kwargs)
        # The wavelength orders are in inverse order in the .json file
        # because I was to lazy to invert them in the file
        wave_range = wave_range[::-1]
        return wave_range



"""
Handles instrument specific info for the UVES spectrograph

Mostly reading data from the header
"""

import logging
import os.path

from .common import Instrument

logger = logging.getLogger(__name__)


class UVES(Instrument):
    def add_header_info(self, header, mode, **kwargs):
        """read data from header and add it as REDUCE keyword back to the header"""
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)

        header["e_ra"] /= 15
        if header["e_jd"] is not None:
            header["e_jd"] += header["e_exptime"] / (7200 * 24) + 0.5

        return header

    def get_wavecal_filename(self, header, mode, **kwargs):
        """Get the filename of the wavelength calibration config file"""
        info = self.load_info()
        specifier = int(header[info["wavecal_specifier"]])

        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_{specifier}nm_2D.npz".format(
            instrument="uves", mode=mode.lower(), specifier=specifier
        )
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname

    def get_mask_filename(self, mode, **kwargs):
        i = self.name.lower()
        m = mode.lower()
        fname = f"mask_{i}_{m}.fits.gz"
        cwd = os.path.dirname(__file__)
        fname = os.path.join(cwd, "..", "masks", fname)
        return fname



"""
Simple script that converts IDL savefiles into numpy savefiles
"""

import os

import numpy as np
from numpy.lib.recfunctions import append_fields
from scipy.io import readsav


def remove_field_name(a, name):
    names = list(a.dtype.names)
    if name in names:
        names.remove(name)
    b = a[names]
    return b


files = [f for f in os.listdir() if f.endswith(".sav")]

if len(files) == 0:
    print("Nothing to convert")

for f in files:
    data = readsav(f)

    cs_lines = data["cs_lines"]
    mask = ~cs_lines.flag.astype(bool)

    cs_lines = remove_field_name(cs_lines, "FLAG")
    cs_lines = append_fields(cs_lines, "flag", mask, usemask=False, asrecarray=True)

    data["cs_lines"] = cs_lines

    new_name = f[:-4]
    np.savez(new_name, **data)
    print(f"Converted {f} to {new_name}.npz")



"""
Handles instrument specific info for the HARPS spectrograph

Mostly reading data from the header
"""

import logging
import os.path

from .common import Instrument

logger = logging.getLogger(__name__)


class JWST_MIRI(Instrument):
    def add_header_info(self, header, mode, **kwargs):
        """read data from header and add it as REDUCE keyword back to the header"""
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)
        self.load_info()
        return header

    def get_wavecal_filename(self, header, mode, **kwargs):
        """Get the filename of the wavelength calibration config file"""
        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_2D.npz".format(instrument="harps", mode=mode)
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname
